
#include "MedianCut.h"
#include "point.h"
#include "Region.h"
//----------------------------------------------------------------------------------------------------------------------
// Load the median cut image
MedianCut::MedianCut(
                     const char * _img,
                     bool& _result
                    ):
                      m_exr(),
                      m_sums( 0 )
{
	// Load the image
  _result = m_exr.load( _img );

	// Get the largest dimension
	int maxDimension = m_exr.width() > m_exr.height() ? m_exr.width() : m_exr.height();

	// Create the temporary array
	m_sums = new float[maxDimension+1];
  if ( !m_sums )
  {
    throw Error( "Failed to allocate sums[]" );
  }
}

//----------------------------------------------------------------------------------------------------------------------
// Delete the array
MedianCut::~MedianCut()
{
	// Delete the sums array
	delete [] m_sums;
}
//----------------------------------------------------------------------------------------------------------------------

// Get the luminance for a colour
float MedianCut::Lum(
                     const Imf::Rgba& _c
                    )
{
  return 0.2125f * _c.r + 0.7154f * _c.g + 0.0721f * _c.b;
	// return ( c.r + c.g + c.b ) / 3.0f;
}

//----------------------------------------------------------------------------------------------------------------------
// Cut the image and generate the light positions
int		MedianCut::cut(
                     int _n,
                     std::vector<MCRegion>& o_regions
                    )
{
	MCRegion r0, r1;

	// Clear the regions
  o_regions.clear();

	// Clear the light list
	m_lights.clear();

	// Create the base region
  o_regions.push_back( MCRegion( 0, m_exr.height(), 0, m_exr.width() ) );
	// While the iterations is less than n, keep going
  while ( (int)o_regions.size() < _n )
	{
		// Subdivide only the regions currently in the list.
    int regionCount = o_regions.size();

		// For each region in the list
		for ( int i = 0; i < regionCount; ++i )
		{
			// Get this region
      MCRegion * r = &o_regions[i];

			// Get the sum of the region
			int count = r->sum( &m_exr, m_sums );

			// Get the median
			int median = MedianCut::Median( m_sums, count );

			// Cut at this position
			r->cut( median, r0, r1 );

			// If this isn't valid, skip it
			if ( !r0.isValid() || !r1.isValid() )
				continue;

			// Set this one as the new region
			*r = r0;

			// Push the new region
      o_regions.push_back( r1 );
		}
	}

	// Return the number of lights
  return (int)o_regions.size();
}

//----------------------------------------------------------------------------------------------------------------------
// Create lights
void MedianCut::createLightsFromRegions(
                                        const std::vector<MCRegion>& _regions
                                       )
{
	// For each region
  std::vector<MCRegion>::const_iterator begin = _regions.begin();
  std::vector<MCRegion>::const_iterator end = _regions.end();

  for ( std::vector<MCRegion>::const_iterator r = begin;r != end; ++r )
	{
		Point p;
		int lx, ly;

		// Get the median point of this box
		r->medianPoint( &m_exr, lx, ly );

		// Save the UV coordinates
    p.m_imgX = lx;
    p.m_imgY = ly;

		// Set the uv coords
    p.m_x = 1 - lx / (float)m_exr.width() - 0.5f;
    p.m_y = 1 - ly / (float)m_exr.height() - 0.5f;

		// Average the pixels in this box
    p.m_colour = sumRegion( *r ); // m_exr.pixel( lx, ly );

		// Add the point
		m_lights.push_back( p );
	}
}

//----------------------------------------------------------------------------------------------------------------------
// Get a light
const Point& MedianCut::getLight(
                                 int _index
                                ) const
{
  if ( _index >= (int)m_lights.size() )
  {
    throw Error( "index out of range for MedianCut::getLight()" );
  }
  return m_lights[_index];
}

//----------------------------------------------------------------------------------------------------------------------
// Get the average of a region
Imf::Rgba MedianCut::sumRegion(
                               const MCRegion& _r
                              ) const
{
	// Imf::Rgba total( 0, 0, 0 );
	float R( 0 ), G( 0 ), B( 0 );

	// Get the area
  float area = 1 / (float)_r.area();
  if ( !area )
  {
    return Imf::Rgba( 0, 0, 0 );
  }
	// For each pixel...
  for ( int x = _r.m_left; x < _r.m_right; ++x )
	{
    for ( int y = _r.m_top; y < _r.m_bottom; ++y )
		{
			// Get the colour at this point
			Imf::Rgba p = m_exr.pixel( x, y );

			// Add this pixel value
			R += p.r;
			G += p.g;
			B += p.b;
		}
	}

	// Divide by the area
	R *= area;
	G *= area;
	B *= area;

	// Return the total
	return Imf::Rgba( R, G, B );
}

//----------------------------------------------------------------------------------------------------------------------
// Write an EXR file of the regions
bool MedianCut::writeRegionMap(
                               const char * _filename,
                               const std::vector<MCRegion>& _regions
                              ) const
{
	// Get the filename
  std::string file( std::string( _filename ) + ".map.exr" );

	// Create a region
	MCRegion mainRegion( 0, m_exr.height(), 0, m_exr.width() );

	// Create the pixels
	Imf::Rgba * pixels = new Imf::Rgba[mainRegion.area()];
	if ( !pixels ) throw Error( "Couldn't allocate memory for pixels in MedianCut::writeRegionMap()" );
  int size=_regions.size();
	// Write each region
  for ( int i = 0; i < size; ++i )
	{
		// Get this region
    const MCRegion * region = &_regions[i];

		// Get this light
		const Point * light = &m_lights[i];

		// Now write the region
    for ( int x = region->m_left; x < region->m_right; ++x )
		{
      for ( int y = region->m_top; y < region->m_bottom; ++y )
			{
				// Get the pixel index
				int index = mainRegion.pixelIndex( x, y );

				// If it's an edge, make it white green
        if ( y == region->m_top || y == region->m_bottom - 1 ||
             x == region->m_left || x == region->m_right - 1 )
				{
					pixels[index] = Imf::Rgba( 0, 0, 0, 1 );
					continue;
				}

				// Print a dot for the light source
        if ( light->m_imgX == x && light->m_imgY == y )
				{
					pixels[index] = Imf::Rgba( 1, 1, 1, 1 );
					continue;
				}

				// Write the pixel
        pixels[index] = light->m_colour;
			}
		}
	}

	// Write the image
	Imf::RgbaOutputFile exr( file.c_str(), mainRegion.width(), mainRegion.height(), Imf::WRITE_RGBA );
	exr.setFrameBuffer( pixels, 1, mainRegion.width() );
	exr.writePixels( mainRegion.height() );

	// Delete the pixels
	delete [] pixels;

	return true;
}

//----------------------------------------------------------------------------------------------------------------------
// Generate a spherical coordinate from an image point
void MedianCut::GetSphericalCoordinateSphere(
                                             float _x,
                                             float _y,
                                             float& o_theta,
                                             float& o_phi
                                            )
{
	// Get the spherical coords
  o_theta = 2 * _x * M_PI;
  o_phi=M_PI-_y*M_PI;

}


//----------------------------------------------------------------------------------------------------------------------
void MedianCut::GetSphericalCoordinateDome(
                                            float _x,
                                            float _y,
                                            float& o_theta,
                                            float& o_phi
                                          )
{
	// Get the spherical coords
  o_theta = 2 * _x * M_PI;
  o_phi   = M_PI/2.0 - _y * M_PI; //PI - y * PI;
}



//----------------------------------------------------------------------------------------------------------------------
// Get the median from some numbers
int MedianCut::Median(
                      const float * _sums,
                      int _count
                      )
{
	// Get the last number and half it
  float median = _sums[_count-1] * 0.5;

	// Find the middle value
	int i;
  for ( i = 0; i < _count - 1; ++i )
	{
		// If it's equal or between this and the next number, return this
    if ( median >= _sums[i] && median < _sums[i+1] )
    {
      return i;
    }
  }
	// Return the cutting index
	return i;
}

//----------------------------------------------------------------------------------------------------------------------
