#include "Region.h"


//----------------------------------------------------------------------------------------------------------------------
// Is the region valid?
bool MCRegion::isValid() const
{
	return width() > 0 && height() > 0;
}

//----------------------------------------------------------------------------------------------------------------------
// Get the dimensions
int MCRegion::width() const { return m_right - m_left; }
//----------------------------------------------------------------------------------------------------------------------
int MCRegion::height() const { return m_bottom - m_top; }

//----------------------------------------------------------------------------------------------------------------------
// Get the middle
void MCRegion::middle(
                    int& o_x,
                    int& o_y
                   ) const
{
  if ( m_left >= m_right - 1 ) o_x = m_left;
  if ( m_top >= m_bottom - 1 ) o_y = m_top;
  o_x = ( m_left + m_right - 1 ) / 2;
  o_y = ( m_top + m_bottom - 1 ) / 2;
}

//----------------------------------------------------------------------------------------------------------------------
// Get a pixel index
int MCRegion::pixelIndex(
                       int _x,
                       int _y
                      ) const
{ 
  return _y * width() + _x;
}

//----------------------------------------------------------------------------------------------------------------------
// Get the area
int MCRegion::area() const
{
	return width() * height();
}

//----------------------------------------------------------------------------------------------------------------------
// Get the largest length
int MCRegion::largestLength() const
{
	return ( width() > height() ) ? width() : height();
}

//----------------------------------------------------------------------------------------------------------------------
// Get the horizontal sum
int MCRegion::horizontalSum(
                          const EXR * _exr,
                          float * _sums
                         ) const
{
	// Reset the first element
  _sums[0] = 0;

	// Scan horizontally
	for ( int x = 0; x < width(); ++x )
	{
		// Copy the sum from the previous column
    if ( x > 0 )
    {
      _sums[x] = _sums[x-1];
    }
		// For each element in the column, add it to the total luminance
    for ( int y = m_top; y < m_bottom; ++y )
    {
      _sums[x] += _exr->lum( x+m_left, y ) * CosineScale( (float)y, (float)_exr->height() );
    }
  }

	// Return the number of boxes filled
	return width();
}

//----------------------------------------------------------------------------------------------------------------------
// Get the vertical sum
int MCRegion::verticalSum(
                        const EXR * _exr,
                        float * _sums
                       ) const
{
	// Reset the first element
  _sums[0] = 0;

	// Scan vertically
	for ( int y = 0; y < height(); ++y )
	{
		// Copy the sum from the previous column
    if ( y > 0 )
    {
      _sums[y] = _sums[y-1];
    }
		// Get the scale
    float scale = CosineScale( (float)(y+m_top), (float)_exr->height() );
		
		// For each element in the column, add it to the total luminance
    for ( int x = m_left; x < m_right; ++x )
    {

      _sums[y] += _exr->lum( x, y+m_top ) * scale;
    }
   }
	
	// return the number of sums
	return height();
}

//----------------------------------------------------------------------------------------------------------------------
// Sum along the longest dimension
int MCRegion::sum(
                const EXR * _exr,
                float * _sums
               ) const
{
	// Sum the longest dimension
  if ( width() > height() )
  {
    return horizontalSum( _exr, _sums );
  }
  else
  {
    return verticalSum( _exr, _sums );
  }
}

//----------------------------------------------------------------------------------------------------------------------
// Find the median point
void MCRegion::medianPoint(
                         const EXR * _img,
                         int& o_x,
                         int& o_y
                        ) const
{
	// Get the width and height
	int hCount = width(), vCount = height();
	
	// Create a list of sums of the largest dimension
	float * hsums = new float[ hCount ];
	float * vsums = new float[ vCount ];
	
	// Reset the first index
	hsums[0] = vsums[0] = 0;
	
	// Sum in both directions
	for ( int h = 0; h < hCount; ++h )
	{
		// Reset
		if ( h > 0 ) hsums[h] = hsums[h-1];
		
		// Now loop through each row
		for ( int v = 0; v < vCount; ++v )
		{
			// Get the luminance for this pixel
      float lum = _img->lum( h + m_left, v + m_top );
			
			// Add to the luminance for this column
			hsums[h] += lum;
			
			// Add to the luminance for this row
			vsums[v] += lum;
		}
	}
	
	// Accumulate the vertical set
	for ( int v = 1; v < vCount; ++v )
	{
		vsums[v] += vsums[v-1];
	}
	
	// Get the medians
  o_x = MedianCut::Median( hsums, hCount ) + m_left;
  o_y = MedianCut::Median( vsums, vCount ) + m_top;

	// Delete the sums
	delete [] hsums;
	delete [] vsums;
}

//----------------------------------------------------------------------------------------------------------------------
// Cut the region
void MCRegion::cut(
                 int _median,
                 MCRegion& o_r0,
                 MCRegion& o_r1
                 ) const
{
	// Copy our properties to the new region
  o_r0 = *this;
  o_r1 = *this;

	// This assumes that you are cutting along the longest dimension
	if ( width() > height() )
	{
		// Cut horizontally
    o_r0.m_right = m_left + _median;
    o_r1.m_left = o_r0.m_right;
	}
	else
	{
		// Cut vertically
    o_r0.m_bottom = m_top + _median;
    o_r1.m_top = o_r0.m_bottom;
	}
}

//----------------------------------------------------------------------------------------------------------------------
// Get the cosine value which maps to
// where max = 1
// c = 0 @ d = 0
// c = 1 @ d = 0.5
// c = 0 @ d = 1
inline float CosineScale(
                         float _d,
                         float _max
                         )
{
  return sinf( _d / _max * M_PI ); // * 0.5 + 0.5;
}

//----------------------------------------------------------------------------------------------------------------------
// Attenuate a colour
void Attenuate(
               Imf::Rgba& _c,
               float _scale
              )
{
  _c.r *= _scale;
  _c.g *= _scale;
  _c.b *= _scale;
}

//----------------------------------------------------------------------------------------------------------------------
// Attenuate a colour
void AttenuateCosine(
                     Imf::Rgba& _c,
                     float _d,
                     float _max
                    )
{
	// Scale it
  Attenuate( _c, CosineScale( _d, _max ) );
}

//----------------------------------------------------------------------------------------------------------------------
// Attenuate a scalar
void AttenuateCosine(
                     float& o_f,
                     float _d,
                     float _max )
{
  o_f *= CosineScale( _d, _max );
}

//----------------------------------------------------------------------------------------------------------------------



